<table class=".origin_table">
    <tr>
        <td style="text-align: left; border-style: hidden">{{ origin_txt }}</td>
        <td style="text-align: right; border-style: hidden">[source]({{ source_url }})</td>
    </tr>
</table>
