
.content-block {
    padding-top: 20px;
    padding-bottom: 10px;
    margin-left: 30px;
    margin-right: 30px;
  }


  @media(min-width: 900px) {
  .content-block {
    margin-left: 50px;
    margin-right: 50px;
  }
  }

  @media (min-width: 1200px) {
  .content-block {
    max-width: 1100px;
    margin-left: auto;
    margin-right: auto;
  }
  }

  .hero-banner {
    position: relative;
    background-color: rgb(237,243,249);
    display: flex;
    justify-content: center;
  }

  .hero-banner h1 {
    color: #39729E;
    font-size: 2.5rem;
  }


  .hero-banner .hero-image {
    position: absolute;
    display: none;
    height: auto;
  }

  @media (min-width: 1000px) {
  .hero-banner .hero-image {
    display: initial;
    width: 270px;
  }
  }

  @media (min-width: 1200px) {
  .hero-banner .hero-image {
    width: 340px;
  }
  }

  @media (min-width: 1400px) {
  .hero-banner .hero-image {
    width: 440px;
  }
  }

  .hero-banner .hero-image p {
    margin-bottom: 0;
  }

  .hero-banner .hero-image-left {
    left: 0;
    bottom: 0;
  }

  .hero-banner .hero-image-right {
    right: 0;
    bottom: 0;
  }


  .hero-banner .content-block {
    max-width: 600px;
    z-index: 2;
  }

  .hero-banner a {
    text-decoration: none;
  }

  .hero-banner h3 {
    margin-top: 1.3rem;
    margin-bottom: 1.3rem;
  }

  .hero-banner h4 {
    margin-top: 0;
  }

  .hero-banner a[role="button"] {
    margin-right: 17px;
    margin-top: 0.6rem;
    margin-bottom: 1.6rem;
  }


  .hero-banner #btn-guide {
    background-color: #959595 !important;
    border: none;
  }




  .hero-banner ul {
    padding-inline-start: 21px;
    font-size: 1.1rem;
  }

  .hero-banner ul li {
    padding-bottom: 0.4rem;
  }


  .alt-background {
    background-color: rgb(247,249,251);
    border-top: 1px solid #dee2e6;
    border-bottom: 1px solid #dee2e6;
  }

  .hello-quarto {
    padding-bottom: 1rem;
  }

  @media (min-width: 600px) {
  .hello-quarto-banner {
    display: inline-flex;
    align-content: center;
    justify-content: center;
  }

  .hello-quarto-banner h1 {
    margin-right: 40px;
  }

  .hello-quarto-banner ul {

  }
  }

  .hello-quarto-banner .nav-pills .nav-link.active, .nav-pills .show>.nav-link {
    border: none;
    border-bottom: 2px solid #39729E !important;
    color: #39729E;
    background-color: transparent;

  }

  .hello-quarto-banner .nav-pills button {
    width: 125px;
  }

  .hello-quarto .tab-content {
    border: none;
    padding: 0;
    color: rgb(84, 85, 85);
  }

  .hello-quarto .tab-content p {
    font-size: 1.1em;
    margin-bottom: 1.5em;
  }

  .hello-quarto div.sourceCode {
    background-color: white;
    border: 1px solid #dee2e6;
  }

  .hello-output {
    background-color: white;
    border: 1px solid #dee2e6;
    max-height: 660px;
  }

  .features {
    padding-bottom: 2em;
  }

  .feature {
    margin-top: 20px;
  }

  @media (min-width: 800px) {
  .features {
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
    margin: 0 0 0 -30px;
    width: calc(100% + 30px);
  }
  .feature {
    width: calc(33% - 30px);
    margin: 20px 0 0 30px;
  }
  }

  .feature h3 {
    margin-top: 0;
  }

  .feature p:first-of-type {
    margin-bottom: 0.2rem;
    color: rgb(84, 85, 85);
  }

  .get-started {
     text-align: center;
     padding-bottom: 2rem;
  }

  .get-started h3 {
     margin-top: 1rem;
     margin-bottom: 2rem;
  }

  nav.page-navigation {
    display: none;
  }

  .nav-footer {
    border-top: none !important;
  }

  .nav-pills .nav-link.active, .nav-pills .show>.nav-link {
    color: #fff;
    background-color: #ff7518;
  }
