"""
Tests for creating/registering accessors.
"""


# class TestDFSBasics:
#     """Test the basic dascore namespace functions."""
#
#     def test_namespace_exists(self, terra15_das_array):
#         """Just ensure namespace returns accessor."""
#         assert hasattr(terra15_das_array, "dascore"), "accessor not registered"
#         acc = terra15_das_array.dascore
#         name = trim_by_time.__name__
#         out = getattr(acc, name)
#         assert callable(out)
